#if !defined(PARSER_HH)
#define PARSER_HH

#include <iostream>

#include "Lexer.hh"
#include "IrManager.hh"
#include "ErrorManager.hh"

///////////////////////////////////////////////////////////////////////////
class Parser {
///////////////////////////////////////////////////////////////////////////

public :

  // Constructor and destructor. The constructor requires:
  // a) a non-const reference to the lexer (to iterate the input stream)
  // b) a non-const reference to the IR manager (to provide IR)
  // c) a non-const reference to the error manager
  Parser( Lexer& lex, IrManager& ir, ErrorManager& error ) : 
    lex_(lex), ir_(ir), error_(error) { }
  ~Parser( void ) { }

  // Main parser: iterates through the list of statements
  AstNode* operator()( void ) {
    AstNode* firstStatement = parseStatement_();
    AstNode* curStatement = firstStatement;
    while (curStatement != 0) {
      AstNode* nextStatement = parseStatement_();
      if (nextStatement != 0) {
	curStatement->setNext(nextStatement);
      }
      curStatement = nextStatement;
    }
    // Check for garbage at the end of the program
    if (lex_.cur().getPartOfSpeech() != Token::EOFF) {
      error_("Garbage after end of program");
    }
    return firstStatement;
  }  
  
private :
  Lexer&        lex_;
  IrManager&    ir_;
  ErrorManager& error_;

  // statement ::= definition | expression
  AstNode* parseStatement_( void ) {
    AstNode* s = 0;
    // Is it an expression or a variable definition?
    // Need to lookahead one token
    lex_.next();
    if (lex_.cur().getPartOfSpeech() == Token::IDFIER) {
      // The statement can only be a variable definition
      s = parseDefinition_();
    } else if ((lex_.cur().getPartOfSpeech() == Token::NUMBER) ||
	       (lex_.cur().getPartOfSpeech() == Token::LPAR)) {
      // The statement can only be an expression evaluation
      s = parseExpression_();
    }
    return s;
  }
  
  // definition ::= identifier = expression
  AstNode* parseDefinition_( void ) {
    AstNode* d = 0;
    // We know that the first token is an identifier
    string identifier = lex_.cur().getLexeme();
    // Now we check that the definition is well formed
    // Here we must encounter an ISEQ token
    lex_.next();
    if (lex_.cur().getPartOfSpeech() != Token::ISEQ) {
      error_("Missing = in definition");
    }
    // Here we must encounter an expression
    // (parseExpression_ assumes that lookahead has been performed)
    lex_.next();
    AstNode* e = parseExpression_();
    if (e == 0) {
      error_("Missing expression in definition");
    }
    // Finally we must find or add the identifier in the symbol table
    // and get its unique numerical identifier
    int uniqueId = ir_.findOrAddId(identifier);
    d = ir_.newOperator(AstNode::ISEQ);
    d->setOperand(AstNode::LEFT, ir_.newVariable(uniqueId));
    d->setOperand(AstNode::RIGHT, e);
    return d;
  }

  // expression ::= ( term op term ) | number
  AstNode* parseExpression_( void ) {
    AstNode* e = 0;
    // We have already looked ahead one token previously
    // so there is no need to advance the token stream here
    if (lex_.cur().getPartOfSpeech() == Token::NUMBER) {
      // Parse a number (a constant)
      e = ir_.newOperand();
      e->setValue(atoi(lex_.cur().getLexeme().data()));
    } else if (lex_.cur().getPartOfSpeech() == Token::LPAR) {
      // Parse a parenthesized expression
      e = ir_.newOperator();
      e->setOperand(AstNode::LEFT, parseTerm_());
      if (e->getOperand(AstNode::LEFT) == 0) {
	error_("Missing expression");
      }
      e->setOpType(parseOpType_());
      if (e->getOpType() == AstNode::NOP) {
	error_("Missing operator");
      }
      e->setOperand(AstNode::RIGHT, parseTerm_());
      if (e->getOperand(AstNode::RIGHT) == 0) {
	error_("Missing expression");
      }
      lex_.next();      
      if (lex_.cur().getPartOfSpeech() != Token::RPAR) {
	error_("Missing )");
      }
    } 
    return e;
  }

  // term ::= expression | identifier
  AstNode* parseTerm_( void ) {
    AstNode* t = 0;
    lex_.next();
    if (lex_.cur().getPartOfSpeech() == Token::IDFIER) {
      // Parse an identifier (a variable): check if it was declared before
      int id = ir_.findId(lex_.cur().getLexeme()); 
      if (id == IrManager::NIL_ID) {
	error_("Undeclared identifier");
      }
      t = ir_.newVariable(id);
    } else {
      // Parse an expression (lookahed of one token is already done)
      t = parseExpression_();
    }
    return t;
  } 

  AstNode::OpType parseOpType_() {
    lex_.next();
    char ch = lex_.cur().getLexeme()[0];
    switch (ch) {
      case '+' : return AstNode::PLUS;
      case '-' : return AstNode::MINUS;
      case '*' : return AstNode::TIMES;
      case '/' : return AstNode::FRAC;
    default :
      return AstNode::NOP;
    }
  }

}; // End of class Parser

#endif
